@@ -154,11 +154,10 @@ def sin(inp):
     elif hasStochastic and isinstance(inp, distribution.stochasticDistribution):
         return distribution.stochasticDistribution(sin(inp.values), inp.probabilities.copy())._update(inp)
     elif not isinstance(inp, oofun): return np.sin(inp)
-    r = oofun(st_sin, inp, 
+    r = oofun(st_sin, inp, engine = 'sin', 
                  d = lambda x: Diag(np.cos(x)), 
                  vectorized = True)
     r._interval_ = lambda domain, dtype: sin_cos_interval(r, inp, domain, dtype)
-    r.repr = lambda *args: 'sin('+ inp.repr() + ')'
     return r
 
 st_cos = (lambda x: \
@@ -175,11 +174,10 @@ def cos(inp):
     elif hasStochastic and isinstance(inp, distribution.stochasticDistribution):
         return distribution.stochasticDistribution(cos(inp.values), inp.probabilities.copy())._update(inp)
     elif not isinstance(inp, oofun): return np.cos(inp)
-    r = oofun(st_cos, inp, 
+    r = oofun(st_cos, inp, engine = 'cos', 
                  d = lambda x: Diag(-np.sin(x)), 
                  vectorized = True)
     r._interval_ = lambda domain, dtype: sin_cos_interval(r, inp, domain, dtype)
-    r.repr = lambda *args: 'cos('+ inp.repr() + ')'
     return r
 #cos = lambda inp: sin(inp+np.pi/2)
 
@@ -207,7 +205,7 @@ def tan(inp):
     if not isinstance(inp, oofun): return np.tan(inp)
     # TODO: move it outside of tan definition
     r = oofun(st_tan, inp, d = lambda x: Diag(1.0 / np.cos(x) ** 2), vectorized = True, \
-    engine_monotonity = 1, convexities = (-1, 1))
+    engine_monotonity = 1, convexities = (-1, 1), engine = 'tan')
     r._interval_ = lambda domain, dtype: tan_interval(inp, r, domain, dtype)
     return r
     
@@ -256,7 +254,7 @@ def arccos(inp):
         return distribution.stochasticDistribution(arccos(inp.values), inp.probabilities.copy())._update(inp)     
     if not isinstance(inp, oofun): return np.arccos(inp)
     r = oofun(st_arccos, inp, d = lambda x: Diag(-1.0 / np.sqrt(1.0 - x**2)), vectorized = True, 
-    engine_monotonity = -1, convexities = (1, -1))
+    engine_monotonity = -1, convexities = (1, -1), engine = 'arccos')
 #    r.getDefiniteRange = get_box1_DefiniteRange
     r._interval_ = lambda domain, dtype: box_1_interval(inp, r, np.arccos, domain, dtype)
     r.attach((inp>-1)('arccos_domain_lower_bound_%d' % r._id, tol=-1e-7), (inp<1)('arccos_domain_upper_bound_%d' % r._id, tol=-1e-7))
@@ -277,7 +275,7 @@ def arctan(inp):
         return distribution.stochasticDistribution(arctan(inp.values), inp.probabilities.copy())._update(inp)             
     if not isinstance(inp, oofun): return np.arctan(inp)
     r = oofun(st_arctan, inp, d = lambda x: Diag(1.0 / (1.0 + x**2)), 
-                 vectorized = True, engine_monotonity = 1, convexities = (1, -1))
+                 vectorized = True, engine_monotonity = 1, convexities = (1, -1), engine = 'arctan')
     r._interval_ = lambda domain, dtype: devided_interval(inp, r, domain, dtype)
     return r
 
@@ -298,7 +296,7 @@ def sinh(inp):
         return distribution.stochasticDistribution(sinh(inp.values), inp.probabilities.copy())._update(inp)        
     if not isinstance(inp, oofun): return np.sinh(inp)
     r = oofun(st_sinh, inp, d = lambda x: Diag(np.cosh(x)), vectorized = True, 
-    engine_monotonity = 1, convexities = (-1, 1))
+    engine_monotonity = 1, convexities = (-1, 1), engine = 'sinh')
     r._interval_ = lambda domain, dtype: devided_interval(inp, r, domain, dtype)
     return r
 
@@ -341,8 +339,9 @@ def cosh(inp):
         return distribution.stochasticDistribution(cosh(inp.values), inp.probabilities.copy())._update(inp)                
     if not isinstance(inp, oofun): 
         return np.cosh(inp)
-    return oofun(st_cosh, inp, d = lambda x: Diag(np.sinh(x)), engine_convexity = 1, vectorized = True, \
-    _interval_=ZeroCriticalPointsInterval(inp, np.cosh))
+    r =oofun(st_cosh, inp, d = lambda x: Diag(np.sinh(x)), engine_convexity = 1, vectorized = True, \
+    _interval_=ZeroCriticalPointsInterval(inp, np.cosh), engine = 'cosh')
+    return r
     
 __all__ += ['sinh', 'cosh']
 
@@ -361,7 +360,7 @@ def tanh(inp):
         return distribution.stochasticDistribution(tanh(inp.values), inp.probabilities.copy())._update(inp)              
     if not isinstance(inp, oofun): return np.tanh(inp)
     r = oofun(st_tanh, inp, d = lambda x: Diag(1.0/np.cosh(x)**2), vectorized = True, 
-    engine_monotonity = 1, convexities = (1, -1))
+    engine_monotonity = 1, convexities = (1, -1), engine = 'tanh')
     r._interval_ = lambda domain, dtype: devided_interval(inp, r, domain, dtype)
     return r
     
@@ -380,7 +379,7 @@ def arctanh(inp):
         return distribution.stochasticDistribution(arctanh(inp.values), inp.probabilities.copy())._update(inp)          
     if not isinstance(inp, oofun): return np.arctanh(inp)
     r = oofun(st_arctanh, inp, d = lambda x: Diag(1.0/(1.0-x**2)), vectorized = True, 
-    engine_monotonity = 1, convexities = (-1, 1))
+    engine_monotonity = 1, convexities = (-1, 1), engine = 'arctanh')
 #    r.getDefiniteRange = get_box1_DefiniteRange
     r._interval_ = lambda domain, dtype: box_1_interval(inp, r, np.arctanh, domain, dtype)
     return r
@@ -402,7 +401,7 @@ def arcsinh(inp):
         return distribution.stochasticDistribution(arcsinh(inp.values), inp.probabilities.copy())._update(inp)      
     if not isinstance(inp, oofun): return np.arcsinh(inp)
     r = oofun(st_arcsinh, inp, d = lambda x: Diag(1.0/np.sqrt(1+x**2)), 
-    vectorized = True, engine_monotonity = 1, convexities = (1, -1))
+    vectorized = True, engine_monotonity = 1, convexities = (1, -1), engine = 'arcsinh')
     r._interval_ = lambda domain, dtype: devided_interval(inp, r, domain, dtype)
     return r
 
@@ -421,7 +420,7 @@ def arccosh(inp):
         return distribution.stochasticDistribution(arccosh(inp.values), inp.probabilities.copy())._update(inp)      
     if not isinstance(inp, oofun): return np.arccosh(inp)
     r = oofun(st_arccosh, inp, d = lambda x: Diag(1.0/np.sqrt(x**2-1.0)), vectorized = True, 
-    engine_monotonity = 1, engine_convexity = -1)
+    engine_monotonity = 1, engine_convexity = -1, engine = 'arccosh')
     F0, shift = 0.0, 1.0
     r._interval_ = lambda domain, dtype: nonnegative_interval(inp, np.arccosh, r.d, domain, dtype, F0, shift)
     return r
@@ -551,7 +550,7 @@ def exp(inp):
         return distribution.stochasticDistribution(exp(inp.values), inp.probabilities.copy())._update(inp)      
     if not isinstance(inp, oofun): return np.exp(inp)
     r = oofun(st_exp, inp, d = lambda x: Diag(np.exp(x)), vectorized = True, 
-    engine_convexity = 1, engine_monotonity = 1)
+    engine_convexity = 1, engine_monotonity = 1, engine = 'exp')
     r._interval_ = lambda domain, dtype: exp_interval(r, inp, domain, dtype)
     return r
 
@@ -574,6 +573,7 @@ def sqrt(inp, attachConstraints = True):
 #    engine_monotonity = 1, engine_convexity = -1)
 #    r._interval_ = lambda domain, dtype: nonnegative_interval(inp, np.sqrt, r.d, domain, dtype, 0.0)
     r = inp ** 0.5
+    r.engine = 'sqrt'
     if attachConstraints: r.attach((inp>0)('sqrt_domain_zero_bound_%d' % r._id, tol=-1e-7))
     return r
 
@@ -594,8 +594,10 @@ def abs(inp):
         return distribution.stochasticDistribution(abs(inp.values), inp.probabilities.copy())._update(inp)      
     if not isinstance(inp, oofun): return np.abs(inp)
     
-    return oofun(st_abs, inp, d = lambda x: Diag(np.sign(x)), vectorized = True, _interval_ = ZeroCriticalPointsInterval(inp, np.abs))
-
+    r = oofun(st_abs, inp, d = lambda x: Diag(np.sign(x)), vectorized = True, 
+    _interval_ = ZeroCriticalPointsInterval(inp, np.abs), engine = 'abs')
+    return r
+    
 __all__ += ['abs']
 
 def get_log_b2_coeffs(L, U, d_l, d_u, c_l, c_u):
@@ -686,7 +688,8 @@ def log(inp):
         return distribution.stochasticDistribution(log(inp.values), inp.probabilities.copy())._update(inp)      
     if not isinstance(inp, oofun): return np.log(inp)
     d = lambda x: Diag(1.0/x)
-    r = oofun(st_log, inp, d = d, vectorized = True, engine_monotonity = 1, engine_convexity = -1)
+    r = oofun(st_log, inp, d = d, vectorized = True, engine_monotonity = 1, engine_convexity = -1, 
+              engine = 'log')
     #r._interval_ = lambda domain, dtype: nonnegative_interval(inp, np.log, r.d, domain, dtype, 0.0)
     r._interval_ = lambda domain, dtype: log_interval(inp, domain, dtype)
     r.attach((inp>1e-300)('log_domain_zero_bound_%d' % r._id, tol=-1e-7))
@@ -713,7 +716,8 @@ def log10(inp):
         return distribution.stochasticDistribution(log10(inp.values), inp.probabilities.copy())._update(inp)              
     if not isinstance(inp, oofun): return np.log10(inp)
     d = lambda x: Diag(INV_LOG_10 / x)
-    r = oofun(st_log10, inp, d = d, vectorized = True, engine_monotonity = 1, engine_convexity = -1)
+    r = oofun(st_log10, inp, d = d, vectorized = True, engine_monotonity = 1, engine_convexity = -1, 
+              engine = 'log10')
     #r._interval_ = lambda domain, dtype: nonnegative_interval(inp, np.log10, r.d, domain, dtype, 0.0)
     r._interval_ = lambda domain, dtype: log10_interval(inp, domain, dtype)
     r.attach((inp>1e-300)('log10_domain_zero_bound_%d' % r._id, tol=-1e-7))
@@ -740,7 +744,8 @@ def log2(inp):
         return distribution.stochasticDistribution(log2(inp.values), inp.probabilities.copy())._update(inp)       
     if not isinstance(inp, oofun): return np.log2(inp)
     d = lambda x: Diag(INV_LOG_2/x)
-    r = oofun(st_log2, inp, d = d, vectorized = True, engine_monotonity = 1, engine_convexity = -1)
+    r = oofun(st_log2, inp, d = d, vectorized = True, engine_monotonity = 1, engine_convexity = -1, 
+              engine = 'log2')
 #    r._interval_ = lambda domain, dtype: nonnegative_interval(inp, np.log2, r.d, domain, dtype, 0.0)
     r._interval_ = lambda domain, dtype: log2_interval(inp, domain, dtype)
     r.attach((inp>1e-300)('log2_domain_zero_bound_%d' % r._id, tol=-1e-7))
@@ -766,7 +771,7 @@ def d_dot(x, y):
 
 def dot(inp1, inp2):
     if not isinstance(inp1, oofun) and not isinstance(inp2, oofun): return np.dot(inp1, inp2)
-    r = oofun(f_dot, [inp1, inp2], d=(lambda x, y: d_dot(x, y), lambda x, y: d_dot(y, x)))
+    r = oofun(f_dot, [inp1, inp2], d=(lambda x, y: d_dot(x, y), lambda x, y: d_dot(y, x)), engine = 'dot')
     r.getOrder = lambda *args, **kwargs: (inp1.getOrder(*args, **kwargs) if isinstance(inp1, oofun) else 0) + (inp2.getOrder(*args, **kwargs) if isinstance(inp2, oofun) else 0)
     #r.isCostly = True
     return r
@@ -778,8 +783,10 @@ def cross_d(x, y):
 def cross(a, b):
     if not isinstance(a, oofun) and not isinstance(b, oofun): return np.cross(a, b)
    
-    r = oofun(np.cross, [a, b], d=(lambda x, y: -cross_d(x, y), lambda x, y: cross_d(y, x)))
-    r.getOrder = lambda *args, **kwargs: (a.getOrder(*args, **kwargs) if isinstance(a, oofun) else 0) + (b.getOrder(*args, **kwargs) if isinstance(b, oofun) else 0)
+    r = oofun(np.cross, [a, b], d=(lambda x, y: -cross_d(x, y), lambda x, y: cross_d(y, x)), engine = 'cross')
+    r.getOrder = lambda *args, **kwargs: \
+    (a.getOrder(*args, **kwargs) if isinstance(a, oofun) else 0)\
+    + (b.getOrder(*args, **kwargs) if isinstance(b, oofun) else 0)
     return r
 
 __all__ += ['dot', 'cross']
@@ -788,7 +795,7 @@ def ceil(inp):
     if isinstance(inp, ooarray) and any(isinstance(elem, oofun) for elem in atleast_1d(inp)):
         return ooarray([ceil(elem) for elem in inp])        
     if not isinstance(inp, oofun): return np.ceil(inp)
-    r = oofun(lambda x: np.ceil(x), inp, vectorized = True, engine_monotonity = 1)
+    r = oofun(lambda x: np.ceil(x), inp, vectorized = True, engine_monotonity = 1, engine = 'ceil')
     r._D = lambda *args, **kwargs: raise_except('derivative for FD ceil is unimplemented yet')
     return r
 
@@ -796,7 +803,7 @@ def floor(inp):
     if isinstance(inp, ooarray) and any(isinstance(elem, oofun) for elem in atleast_1d(inp)):
         return ooarray([floor(elem) for elem in inp])        
     if not isinstance(inp, oofun): return np.floor(inp)
-    r = oofun(lambda x: np.floor(x), inp, vectorized = True, engine_monotonity = 1)
+    r = oofun(lambda x: np.floor(x), inp, vectorized = True, engine_monotonity = 1, engine = 'floor')
     r._D = lambda *args, **kwargs: raise_except('derivative for FD floor is unimplemented yet')
     return r
 
@@ -815,7 +822,7 @@ def sign(inp):
         return distribution.stochasticDistribution(sign(inp.values), inp.probabilities.copy())._update(inp)      
     if not isinstance(inp, oofun): 
         return np.sign(inp)
-    r = oofun(st_sign, inp, vectorized = True, d = lambda x: 0.0, engine_monotonity = 1)
+    r = oofun(st_sign, inp, vectorized = True, d = lambda x: 0.0, engine_monotonity = 1, engine = 'sign')
     return r
 
 __all__ += ['ceil', 'floor', 'sign']
@@ -1176,18 +1183,19 @@ def sum(inp, *args, **kwargs):
         r_dep = r._getDep()
         r._D = lambda *args, **kw: sum_derivative(r, r0, INP, r_dep, *args, **kw)
         r.isCostly = True
-        def repr(*args):
-            Elems = [elem.repr() if isinstance(elem, oofun) else str(elem) for elem in INP]
+        
+        def expression(*args, **kw):
+            Elems = [elem.expression(**kw) if isinstance(elem, oofun) else str(elem) for elem in INP]
             r = []
-            for i, elem in enumerate(Elems):
-                r.append(elem +  ('+' if elem[0] != '-' and i != len(Elems)-1 else ''))
-            rr = ''.join(r)
+            for i in range(len(Elems)):
+                r.append(Elems[i] +  (' + ' if i != len(Elems)-1 and Elems[i+1][0] != '-' else ' '))
+            rr = ''.join(r)[:-1]
+            if not np.array_equal(r0, 0.0):
+                str_r0 = str(r0)
+                rr += ' + ' + str_r0 if str_r0[0] != '-' else ' - ' + str_r0[1:]
             return rr
-        r.repr = repr
-#        r.repr = lambda *args: \
-#        ''.join(elem.repr() if isinstance(elem, oofun) else str(elem) + ('+' if i != len(INP)-1 else '') for i, elem\
-#                in enumerate(INP) ) + ('+' + str(r0) if not np.array_equal(r0, 0.0) else '')
-            
+        r.expression = expression
+        
         return r
     else: 
         return inp.sum(*args, **kwargs)#np.sum(inp, *args, **kwargs)
@@ -1204,10 +1212,12 @@ def prod(inp, *args, **kwargs):
 def norm(*args, **kwargs):
     if len(kwargs) or len(args) > 1:
         return np.linalg.norm(*args, **kwargs)
-    r = sqrt(sum(args[0]**2),  attachConstraints=False)
     if isinstance(r, oofun) and len(kwargs) == 0 and (len(args) == 1 or args[1] == 2):
+        r = sqrt(sum(args[0]**2),  attachConstraints=False)
         r.engine = 'norm2'
         r._norm_arg = args[0]#used in SOCP
+    else:
+        assert 0, 'FuncDesigner norm() is implemented for n = 2 only'
     return r
  
 __all__ += ['sum', 'prod', 'norm']
@@ -1273,7 +1283,7 @@ def ifThenElse(condition, val1, val2, *args, **kwargs):
     elif isinstance(condition, oofun):
         f = lambda conditionResult, value1Result, value2Result: value1Result if conditionResult else value2Result
         # !!! Don't modify it elseware function will evaluate both expressions despite of condition value 
-        r = oofun(f, [condition, val1, val2])
+        r = oofun(f, [condition, val1, val2], engine = 'ifThenElse')
         r.D = lambda point, *args, **kwargs: (Val1.D(point, *args, **kwargs) if isinstance(Val1, oofun) else {}) if condition(point) else \
         (Val2.D(point, *args, **kwargs) if isinstance(Val2, oofun) else {})
         r._D = lambda point, *args, **kwargs: (Val1._D(point, *args, **kwargs) if isinstance(Val1, oofun) else {}) if condition(point) else \
@@ -1334,7 +1344,7 @@ def max(inp,  *args,  **kwargs):
             r1[r1<tmp] = tmp
             r2[r2<tmp] = tmp
             return np.vstack((r1, r2)), DefiniteRange
-        r = oofun(f, inp, size = 1, _interval_ = interval)
+        r = oofun(f, inp, size = 1, _interval_ = interval, engine = 'max')
         def _D(point, *args, **kwargs):
             ind = np.argmax([(s(point) if isinstance(s, oofun) else s) for s in r.input])
             return r.input[ind]._D(point, *args, **kwargs) if isinstance(r.input[ind], oofun) else {}
@@ -1381,7 +1391,7 @@ def min(inp,  *args,  **kwargs):
                 r2[r2>tmp] = tmp
             return np.vstack((r1, r2)), DefiniteRange
             
-        r = oofun(f, inp, size = 1, _interval_ = interval)
+        r = oofun(f, inp, size = 1, _interval_ = interval, engine = 'min')
         def _D(point, *args, **kwargs):
             ind = np.argmin([(s(point) if isinstance(s, oofun) else s) for s in r.input])
             return r.input[ind]._D(point, *args, **kwargs) if isinstance(r.input[ind], oofun) else {}
@@ -1444,7 +1454,7 @@ def hstack(tup): # overload for oofun[ind]
         return np.max(orders)
     
             
-    r = oofun(f, tup, getOrder = getOrder)
+    r = oofun(f, tup, getOrder = getOrder, engine = 'hstack')
     
     #!!!!!!!!!!!!!!!!! TODO: sparse 
 
